#!/usr/bin/env python3

from collections.abc import Callable
from typing import Union, TypeVar

T = TypeVar('T')


def leq(a: Union[float, int], b: Union[float, int]) -> bool:
    return a <= b


def geq(a: Union[float, int], b: Union[float, int]) -> bool:
    return a >= b


class BinHeap(object):
    ''' An array-based implementation of binary heaps

    This class provides an array-based implementation of binary heaps.

    Parameters
    ----------
    A : list[T]
        A list of the keys to be inserted in the heap itself.
    total_order : Callable[[T, T], bool]
        The total order of the binary heap.

    Members
    -------
    _keys : list[T]
        A list that stores the binary heap keys.
    _size : int
        The size of the heap.
    _total_order : Callable[[T, T], bool]
        The total order of the binary heap.
    '''
    def __init__(self, A: list[T],
                 total_order: Callable[[T, T], bool] = leq) -> None:
        self._keys = A
        self._size = len(A)
        self._total_order = total_order

        # build heap procedure
        for node in range(self._size-1, -1, -1):
            self._heapify(node)

    @staticmethod
    def left(node: int) -> int:
        return 2*node+1

    @staticmethod
    def right(node: int) -> int:
        return 2*node+2

    @staticmethod
    def parent(node: int) -> int:
        ''' Return the node's parent

        Parameters
        ----------
        node: int
            Node of which we are searching the parent

        Returns
        -------
        int
            The parent of `node`, if it exists; the node itself, otherwise.
        '''
        if node == 0:
            return 0
        return (node-1)//2

    def __len__(self):
        return self._size

    def is_valid(self, node: int) -> bool:
        # a node is valid if it is a valid index for the
        # keys array and it is lesser than the number
        # of nodes in the binary tree
        return 0 <= node < self._size

    def min(self) -> T:
        # if the heap is empty, then raise an exception
        if self._size == 0:
            raise RuntimeError("The binary heap is empty")

        # otherwise, the minimum among all the values in
        # the heap is stored on the root
        return self._keys[0]

    def _swap_keys(self, node_keys: int, node_b: int) -> None:
        (self._keys[node_keys], self._keys[node_b]) = (self._keys[node_b],
                                                       self._keys[node_keys])

    def _heapify(self, node: int) -> None:
        while True:
            # min_node stores the candidate having the minimum
            # key among node and its children. It is initially
            # set to the node itself
            min_node = node

            # for all the possible children of node
            for child in [BinHeap.right(node),
                          BinHeap.left(node)]:

                # if child is a valid node in the binary tree
                # and its key is smaller that that of the candidate
                if (self.is_valid(child) and
                        not self._total_order(self._keys[min_node],
                                              self._keys[child])):
                    # we have discovered a new candidate
                    min_node = child

            # after the loop min_node contains the node
            # whose key is the minimal among those of
            # node and its children

            # if min_node is the node itself, the heap
            # property has been fixed everywhere and
            # the computation can end
            if min_node == node:
                return

            # otherwise, swap the keys of the two nodes
            # and, possibly, push the issue one level down.
            # Then, focus on the node potentially breaking
            # the heap property, i.e., min_node
            self._swap_keys(node, min_node)
            node = min_node

    def extract_min(self):
        # get the minimum by using min()
        min_value = self.min()

        # replace the root key by using the key
        # of the rightmost leaf on the last level
        # and decrease the number of values in
        # the tree by one
        self._keys[0] = self._keys[self._size-1]
        self._size -= 1

        # the reassignament of the root key may
        # have broken the heap property, but only
        # on the root of the binary tree. Thus,
        # we can call heapify to fix it.
        self._heapify(0)

        return min_value

    def decrease_key(self, node: int, new_value: T):
        # if the node is not a binary heap node,
        # raise an exception
        if not self.is_valid(node):
            raise RuntimeError("The node does not exist")

        # if the new value is "greater" than that already
        # in the node, raise an exception
        if not self._total_order(new_value, self._keys[node]):
            raise RuntimeError("The new value is \"greater\" " +
                               "than previous one.")

        # otherwise the node belongs to the binary heap and
        # the new value is lesser than or equal to the
        # old key of the node

        # assign the new value to the node
        self._keys[node] = new_value

        # if the node is the tree root, heap property
        # is already satisfied by all the tree nodes
        # and we can end the procedure
        # otherwise, while the node is not the root
        while node != 0:

            # check whether the heap property between the node and
            # its parent is satisfied
            parent = BinHeap.parent(node)
            if self._total_order(self._keys[parent], self._keys[node]):

                # if this is the case, the heap property is
                # satisfied along all the tree and nothing
                # more is needed
                return

            # if the node and its parent do not satisfy the
            # heap property, swap their keys
            self._swap_keys(parent, node)

            # the swap has fixed the heap property between
            # the node and its parent, but it has decreased
            # the value of the parent keys to, thus, repeat
            # every thing focusing on the parent
            node = parent

    def insert(self, value: T):
        if len(self._keys) == self._size:
            raise RuntimeError("The maximum number of values in the heap " +
                               "has been already reached.")

        # now self._size is a valid index in A
        # and it is the correct position to insert
        # the new node
        node = self._size
        self._size += 1

        # the topological property. We still need to update its value
        # preserving the heap property.

        # if the new node is the root
        if node == 0:
            # the binary heap was empty and we are inserting the
            # root's key. Noting more is needed to conclude the
            # procedure
            self._keys[0] = value
            return

        # we temporary assign the key of the new node to the
        # maximum among the new value and the parent's key
        # so to satisfy the heap property and get a valid
        # binary heap
        parent = BinHeap.parent(node)
        if self._total_order(self._keys[parent], value):

            # if the maximum is the new value, then
            # heap property is already satisfied and
            # we do not need to fix it
            self._keys[node] = value
        else:

            # if the maximum is the parent's key,
            # we can decrease it to the new value
            self._keys[node] = self._keys[parent]
            self.decrease_key(node, value)

    def __repr__(self) -> str:
        missing_on_level = 1
        node_on_level = 1
        result = ""

        for node in range(self._size):
            result += f"{self._keys[node]} "
            missing_on_level -= 1

            if missing_on_level == 0:
                result += "\n"
                node_on_level *= 2
                missing_on_level = node_on_level
        return result


if __name__ == '__main__':
    A = [0, 3, 2, 1, 6, 5, 8, 2]

    H = BinHeap(A, leq)

    print(H)

    H.extract_min()
    H.extract_min()

    print(f"min value = {H.extract_min()}")

    print(H)
    print()
    H.decrease_key(1, -1)

    print(H)
    print()

    H.insert(-4)
    H.insert(4)
    H.insert(2)

    print(H)
    print()

    H.insert(-4)

    print(H)
    print()

